; DLCLD.ASM - binary program loader for EVM56K via SCI port
;             includes code to program flash eeprom
;------------------------------------------------------------------------------
; 26.9.1996 Juergen Hasch, DG1SCR
;                          hasch@t-online.de
;
; Derived from Motorola sample programs, improvements by J. Forrer
;
;------------------------------------------------------------------------------
; Usage:
;  Load dlcld.cld with evm56k debugger, start with GO 40 .
;  Tested with an AT29C256 eeprom.
;
; List of commands:
; -----------------------
; 01 - write block to P memory
; 02 - write block to X memory
; 03 - write block to Y memory
; 04 - execute program
;
; block transfer:
;  1. get start address (16 Bit)
;  2. get block length (16 bit)
;  3. get requested # of 24 bit words
;  4. send CRC checksum (calculated from receive data)
;
	PAGE    132,60,1,1

BAUD    EQU     38400           ; download baudrate
XTAL    EQU     40000000        ; internal clock rate
SCR     EQU     $FFF0           ; SCI CONTROL REGISTER
SCCR    EQU     $FFF2           ; SCI CLOCK CONTROL REGISTER
PCC     EQU     $FFE1           ; PORT C CONTROL REGISTER
PLL     EQU     $FFFD           ; PLL CONTROL REGISTER
BCR     EQU     $FFFE           ; Bus Control Register
RDRF    EQU     $2              ; RECEIVE DATA REGISTER FULL FLAG
RDRX    EQU     $1              ; TRANSMIT DATA REGISTER EMPTY FLAG
SSR     EQU     $FFF1           ; SCI STATUS REGISTER
SRXH    EQU     $FFF6           ; SCI RECEIVE IN HIGH BYTE
SRXM    EQU     $FFF5           ; SCI RECEIVE IN MIDDLE BYTE
SRXL    EQU     $FFF4           ; SCI RECEIVE IN LOW BYTE
poly    EQU     $8408           ; CRC polynomial (x^16 + x^12 + x^5 + 1)
boot    EQU     $3000
start   EQU     $3F00           ; The loader will be copied to p:$7fe0

        org     p:$0040

main
	movep   #$261009,x:PLL          ;set PLL for MF=10, 40 MHz VCO out
	movep   #$B0B0,x:BCR            ;at 40MHz, T=25ns, 11 wait states in P:

;------ move the bootstrap loader into FLASH at P:$C000
	move    #boot,r0               ;first DSP address to save
	move    #$C000,r2               ;flash starting address
	move    #5,r1
	do      r1,loop
	jsr     WR_FLASH                ;save 64 words
loop

	movep  #$0000,x:BCR
	debug
	jmp    *

;************************************************************************
;   WR_FLASH
;   This routine writes 64 (24-bit) words from DSP RAM to FLASH EEPROM
;
;       enter with:     r2 pointing to starting EEPROM Byte Address
;                       r0 pointing to first DSP word to save
;
;-------------------------------------------------------------------------
;   first, move 64 bytes - this includes 21 words + low byte of 22nd
;-------------------------------------------------------------------------
WR_FLASH
	jsr     disable_protect

	do      #21,move_codeA          ;move 63 bytes
	move    P:(r0)+,A               ;get 3 byte word
	move            A1,x:(r2)+      ;move low  byte to FLASH EEPROM
	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move mid  byte to FLASH EEPROM
	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move high byte
move_codeA
					;*** 64th byte
	move    P:(r0)+,A               ;get next 3 byte word
	move            A1,x:(r2)+      ;move low  byte

	jsr     delay_200u              ;wait 200 microseconds to enter
					;...program cycle
	jsr     program_complete        ;wait for programming cycle to
					;...finish
;-------------------------------------------------------------------------
;   then, move another 64 bytes - mid and high bytes of 22nd word followed
;                                 by 20 words and then the low and mid
;                                 bytes of the next word
;-------------------------------------------------------------------------
	jsr     disable_protect

	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move mid  byte to FLASH EEPROM
	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move high byte


	do      #20,move_codeB          ;move 60 bytes
	move    P:(r0)+,A               ;get 3 byte word
	move            A1,x:(r2)+      ;move low  byte to FLASH EEPROM
	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move mid  byte to FLASH EEPROM
	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move high byte
move_codeB
					;*** 63rd byte ***
	move    P:(r0)+,A               ;get next 3 byte word
	move            A1,x:(r2)+      ;move low byte
	rep     #8
	lsr     A                       ;*** 64th byte ***
	move            A1,x:(r2)+      ;move mid  byte to FLASH EEPROM

	jsr     delay_200u              ;wait 200 microseconds to enter
					;...program cycle
	jsr     program_complete        ;wait for programming cycle to




;-------------------------------------------------------------------------
;   lastly, move another 64 bytes - high bytes of last word followed
;                                 by 21 words
;-------------------------------------------------------------------------
	jsr     disable_protect

	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move high byte of last word


	do      #21,move_codeC          ;move 63 bytes
	move    P:(r0)+,A               ;get 3 byte word
	move            A1,x:(r2)+      ;move low  byte to FLASH EEPROM
	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move mid  byte to FLASH EEPROM
	rep     #8
	lsr     A
	move            A1,x:(r2)+      ;move high byte
move_codeC

	jsr     delay_200u              ;wait 200 microseconds to enter
					;...program cycle
	jsr     program_complete        ;wait for programming cycle to
					;...finish

	jsr     enable_protect
	rts
;***************************************************************************

;----------------------------------------------------------------------------
;  enable_protect - Software data Protection routine
;
;----------------------------------------------------------------------------
enable_protect
	move    #$AA,x0
	move            x0,x:$D555   ;Atmel needs $5555
	move    #$55,x0
	move            x0,x:$AAAA   ;Atmel needs $2AAA
	move    #$A0,x0
	move            x0,x:$D555   ;writes are now enabled
	rts

;---------------------------------------------------------------------------
;  disable_protect - Disables Software Data protection
;
;  This code writes the appropriate data to the FLASH part that disables
;  the write protection from the FLASH and allows the part to be loaded.
;
;---------------------------------------------------------------------------
disable_protect
	move    #$AA,x0
	move            x0,x:$D555
	move    #$55,x0
	move            x0,x:$AAAA
	move    #$80,x0
	move            x0,x:$D555
	move    #$AA,x0
	move            x0,x:$D555
	move    #$55,x0
	move            x0,x:$AAAA
	move    #$20,x0
	move            x0,x:$D555      ;exit data protect state
	rts

;---------------------------------------------------------------------------
; delay routine to be used to satisfy tBLC time requirement which is
; a minimum of 150 microseconds.
; Write Pulse Width, Write Pulse Width High requires a minimum of
; 3000 Icycles for 40MHz clock.  (50ns/Icycle)
;---------------------------------------------------------------------------
delay_200u
	rep    #4000      ; delay of 200 micro seconds.
	nop
	rts

;---------------------------------------------------------------------------
;  program_complete - routine to poll for the end of the programming cycle
;
;  The AT29C256 will toggle bit 6 on successive reads until the program
;  cycle completes.  When bit 6 stabilizes, the device is no longer in
;  a programming cycle.
;---------------------------------------------------------------------------
program_complete
	btst    #6,x:$8000              ;check the initial status of bit 6
	jcs     check_set
check_clr                               ;---go here if bit 6 is clear---
	jsr     delay_200u              ;no need to hurry, give it a rest
	btst    #6,x:$8000              ;test bit 6...it was clear
	jcs     check_set               ;if it is now set, continue...
	rts
check_set                               ;---go here if bit 6 is set---
	jsr     delay_200u
	btst    #6,x:$8000              ;test bit 6...it was set
	jcc     check_clr               ;if it is now clear, continue...
	rts


;**************************************************************************
;  this is the bootstrap loader
;  it is loaded into the memory at boot time and then moves the
;  application code into DSP RAM
;
	org     p:0,p:boot
;----------------------------------------------------------------------------
;The following code moves the loader code from the P:RAM.
;----------------------------------------------------------------------------
begin   move    #reloc,r0               ; R0 points to loader code.
        move    #start,r1               ; R1 points to runtime loader loc.
	do      #load_length,_move_code ; Move loader code to runtime loc.
	move    p:(r0)+,x0              ; Read loader code from load    loc.
	move    x0,p:(r1)+              ; Write loader code to  runtime loc.
_move_code
;----------------------------------------------------------------------------
;The following code moves the user's code from the flash to P:RAM.
;----------------------------------------------------------------------------
	jmp     load

reloc
        org     p:start,p:

; byte CRC calculation routine
; byte in a1, result in r6
CRCBYTE macro
	tfr	a,b	#>1,x1
	move                #>poly,y1
	do      #8,_crc2

	move    b1,a1                               ; first LSB to remainder
	and     x1,a        r6,y0
	eor     y0,a

	lsr     a                                   ; XOR if needed
	jcc     _crc1
	eor     y1,a
_crc1   lsr     b           a1,r6
_crc2
	endm


load
	movep   #$261009,X:PLL           ;SET EVM TO 40 MHZ OPERATION
	movep   #$0302,X:SCR             ;ENABLE TX,RX: 8 BIT 1 START, 1 STOP
        movep   #(XTAL+2*16*BAUD)/(2*2*16*BAUD)-1,X:SCCR ; set baudrate
	movep   #$0007,X:PCC             ; ENABLE SCI

WAITCMD
	jsr     GETCH                    ; INPUT CHARACTER

	;
	; check for command
	; 01 - write block to P memory
	; 02 - write block to X memory
	; 03 - write block to Y memory
	; 04 - execute program
	;
	move    #01,X0
	cmp     X0,A
	jeq     WRITEP
	move    #02,X0
	cmp     X0,A
	jeq     WRITEX
	move    #03,X0
	cmp     X0,A
	jeq     WRITEY
	move    #04,X0
	cmp     X0,A
	jne     WAITCMD

	;
	; execute program
	;
EXECUTE
	jsr     PUTCH   ; echo command
	JSR     GET16   ; get 16 bit value in B (exec address)
	move    B1,R0
	nop
	jmp     (R0)

	;
	; write block to program memory
	;
WRITEP
	jsr     PUTCH           ; echo command
	jsr     GET16           ; get start address
	move    A1,R0
	jsr     GET16           ; get length
	move    A1,R1
	clr     A
	move    #0,R6           ; preset CRC

	; loop and write
WPL
	jsr     GET24
	move    A1,P:(R0)+
	clr     A    (R1)-
	nop
	move    R1,B
	tst     B
	jne     WPL

	move    R6,A
	jsr     PUT16           ; send CRC

	jmp     WAITCMD

	;
	; write block to X memory
	;
WRITEX
	jsr     PUTCH           ; echo command
	jsr     GET16           ; get start address
	move    A1,R0
	jsr     GET16           ; get length
	move    A1,R1
	clr     A
	move    #0,R6           ; clear CRC
	; loop and write
WXL
	jsr     GET24
	move    A1,X:(R0)+
	clr     A       (R1)-
	nop
	move    R1,B
	tst     B
	jne     WXL

	move    R6,A
	jsr     PUT16           ; send CRC

	jmp     WAITCMD

	;
	; write block to Y memory
	;
WRITEY
	jsr     PUTCH           ; echo command
	jsr     GET16           ; get start address
	move    A1,R0
	jsr     GET16           ; get length
	move    A1,R1
	clr     A
	move    #0,R6           ; clear CRC
	; loop and write
WYL
	jsr     GET24
	move    A1,Y:(R0)+
	clr     A       (R1)-
	nop
	move    R1,B
	tst     B
	jne     WYL

	move    R6,A
	jsr     PUT16           ; send CRC

	jmp     WAITCMD

;----------------------------------
	;
	; get 16 bit word
	;
GET16
	jclr    #RDRF,X:SSR,*   ; WAIT FOR DATA IN SCI
	movep   X:SRXM,A        ; GET SCI DATA IN HIGH BYTE
	move    A1,R5           ; SIGN EXTEND AND ZERO
	jclr    #RDRF,X:SSR,*   ; WAIT FOR DATA IN SCI
	movep   X:SRXL,A        ; GET SCI DATA IN HIGH BYTE
	move    A1,X0           ; SIGN EXTEND AND ZERO
	move    R5,A1
	or      X0,A1
	rts

	;
	; get 24 bit word
	;
GET24
	jclr    #RDRF,X:SSR,*   ; WAIT FOR DATA IN SCI
	movep   X:SRXM,A        ; GET SCI DATA IN HIGH BYTE
	move    A1,R5           ; SIGN EXTEND AND ZERO
	rep     #8
	lsr     A
	CRCBYTE                 ; update CRC
	jclr    #RDRF,X:SSR,*   ; WAIT FOR DATA IN SCI
	movep   X:SRXL,A        ; GET SCI DATA IN HIGH BYTE
	move    A1,R4           ; SIGN EXTEND AND ZERO
	CRCBYTE                 ; update CRC
	jclr    #RDRF,X:SSR,*   ; WAIT FOR DATA IN SCI
	movep   X:SRXL,A        ; GET SCI DATA IN HIGH BYTE
	move    A1,X0           ; SIGN EXTEND AND ZERO
	CRCBYTE                 ; update CRC
	move    R5,A1
	move    R4,X1
	or      X1,A1
	rep     #8
	lsl     A1
	or      X0,A1
	rts

	;
	; get character
	;
GETCH
	jclr    #RDRF,X:SSR,*   ;WAIT FOR DATA IN SCI
	movep   X:SRXH,A        ;GET SCI DATA IN HIGH BYTE
	move    A1,X0           ;SIGN EXTEND AND ZERO
	move    X0,A
	rts

	;
	; send 16 bit word
	;
PUT16
	jclr    #RDRX,X:SSR,*   ;WAIT UNTIL TRANSMIT REGISTER EMPTY
	movep   A,X:SRXM        ;SEND PROMPT
	jclr    #RDRX,X:SSR,*   ;WAIT UNTIL TRANSMIT REGISTER EMPTY
	movep   A,X:SRXL        ;SEND PROMPT
	RTS
	;
	; send character
	;
PUTCH
	jclr    #RDRX,X:SSR,*   ;WAIT UNTIL TRANSMIT REGISTER EMPTY
	movep   A,X:SRXH        ;SEND PROMPT
	rts

load_length     equ     *-load

	end